﻿/*	see copyright notice in squirrel.h */

import std.c.stdio;
import std.c.stdlib;
import std.c.string;
import std.c.stdarg;
import squirrel;
import std.string : join;

int main(char[][] args)
{
	HSQUIRRELVM v;
	
	const SQChar *filename;
	
	v=sq_open(1024);
	sq_setprintfunc(v,&printfunc);

	sq_pushroottable(v);

	sqstd_register_bloblib(v);
	sqstd_register_iolib(v);
	sqstd_register_systemlib(v);
	sqstd_register_mathlib(v);
	sqstd_register_stringlib(v);

	//aux library
	//sets error handlers
	sqstd_seterrorhandlers(v);

	//gets arguments
	switch(getargs(v,args))
	{
	case _INTERACTIVE:
		Interactive(v);
		break;
	case _DONE:
	default: 
		break;
	}

	sq_close(v);
	
	return 0;
}

extern(C):

alias fprintf scfprintf;
alias fopen scfopen;
alias vprintf scvprintf;


SQInteger quit(HSQUIRRELVM v)
{
	int *done;
	sq_getuserpointer(v,-1,cast(SQUserPointer*)&done);
	*done=1;
	return 0;
}

void printfunc(HSQUIRRELVM v,SQChar *s,...)
{
	va_list vl;
	va_start(vl, s);
	scvprintf( s, vl);
	va_end(vl);
}

void PrintVersionInfos()
{
	scfprintf(stdout,_SC("%.*s %.*s (%d bits)\n"),SQUIRREL_VERSION,SQUIRREL_COPYRIGHT,SQInteger.sizeof*8);
}

void PrintUsage()
{
	scfprintf(stderr,
		"usage: sq <options> <scriptpath [args]>.\n"
		"Available options are:\n"
		"   -c              compiles the file to bytecode(default output 'out.cnut')\n"
		"   -o              specifies output file for the -c option\n"
		"   -c              compiles only\n"
		"   -d              generates debug infos\n"
		"   -v              displays version infos\n"
		"   -h              prints help\n");
}

const _INTERACTIVE = 0;
const _DONE = 2;
//<<FIXME>> this func is a mess
int getargs(HSQUIRRELVM v,char[][] argv)
{
	int argc = argv.length;
	int i;
	int compiles_only = 0;
	static SQChar temp[500];
	const SQChar *ret;
	char * output;
	int lineinfo=0;
	if(argc>1)
	{
		int arg=1,exitloop=0;
		while(arg < argc && !exitloop)
		{

			if(argv[arg][0]=='-')
			{
				switch(argv[arg][1])
				{
				case 'd': //DEBUG(debug infos)
					sq_enabledebuginfo(v,1);
					break;
				case 'c':
					compiles_only = 1;
					break;
				case 'o':
					if(arg < argc) {
						arg++;
						output = argv[arg].ptr;
					}
					break;
				case 'v':
					PrintVersionInfos();
					return _DONE;
				
				case 'h':
					PrintVersionInfos();
					PrintUsage();
					return _DONE;
				default:
					PrintVersionInfos();
					scprintf(_SC("unknown prameter '-%c'\n"),argv[arg][1]);
					PrintUsage();
					return _DONE;
				}
			}else break;
			arg++;
		}

		// src file
		
		if(arg<argc) {
			SQChar *filename=argv[arg].ptr;

			arg++;
			sq_pushroottable(v);
			sq_pushstring(v,_SC("ARGS"),-1);
			sq_newarray(v,0);
			for(i=arg;i<argc;i++)
			{
				SQChar *a=argv[i].ptr;
				sq_pushstring(v,a,-1);

				sq_arrayappend(v,-2);
			}
			sq_createslot(v,-3);
			sq_pop(v,1);
			if(compiles_only) {
				if(SQ_SUCCEEDED(sqstd_loadfile(v,filename,SQTrue))){
					SQChar *outfile = _SC("out.cnut");
					if(output) {
						outfile = output;

					}
					if(SQ_SUCCEEDED(sqstd_writeclosuretofile(v,outfile)))
						return _DONE;
				}
			}
			else {
				if(SQ_SUCCEEDED(sqstd_dofile(v,filename,SQFalse,SQTrue))) {
					return _DONE;
				}
			}
			//if this point is reached an error occured
			{
				SQChar *err;
				sq_getlasterror(v);
				if(SQ_SUCCEEDED(sq_getstring(v,-1,&err))) {
					scprintf(_SC("Error [%s]\n"),err);
					return _DONE;
				}
			}
			
		}
	}

	return _INTERACTIVE;
}

void Interactive(HSQUIRRELVM v)
{
	
const MAXINPUT = 1024;
	SQChar buffer[MAXINPUT];
	SQInteger blocks =0;
	SQInteger string=0;
	SQInteger retval=0;
	SQInteger done=0;
	PrintVersionInfos();
		
	sq_pushroottable(v);
	sq_pushstring(v,_SC("quit"),-1);
	sq_pushuserpointer(v,&done);
	sq_newclosure(v,&quit,1);
	sq_setparamscheck(v,1,null);
	sq_createslot(v,-3);
	sq_pop(v,1);

    while (!done) 
	{
		SQInteger i = 0;
		scprintf(_SC("\nsq>"));
		for(;;) {
			int c;
			if(done)return;
			c = getchar();
			if (c == _SC("\n")[0]) {
				if (i>0 && buffer[i-1] == _SC("\\")[0])
				{
					buffer[i-1] = _SC("\n")[0];
				}
				else if(blocks==0)break;
				buffer[i++] = _SC("\n")[0];
			}
			else if (c==_SC("}")[0]) {blocks--; buffer[i++] = cast(SQChar)c;}
			else if(c==_SC("{")[0] && !string){
					blocks++;
					buffer[i++] = cast(SQChar)c;
			}
			else if(c==_SC("\"")[0] || c==_SC("'")[0]){
					string=!string;
					buffer[i++] = cast(SQChar)c;
			}
			else if (i >= MAXINPUT-1) {
				scfprintf(stderr, _SC("sq : input line too long\n"));
				break;
			}
			else{
				buffer[i++] = cast(SQChar)c;
			}
		}
		buffer[i] = _SC("\0")[0];
		
		if(buffer[0]==_SC("=")[0]){
			scsprintf(sq_getscratchpad(v,MAXINPUT),_SC("return (%s)"),&buffer[1]);
			memcpy(buffer.ptr,sq_getscratchpad(v,-1),(scstrlen(sq_getscratchpad(v,-1))+1)*SQChar.sizeof);
			retval=1;
		}
		i=scstrlen(buffer.ptr);
		if(i>0){
			SQInteger oldtop=sq_gettop(v);
			if(SQ_SUCCEEDED(sq_compilebuffer(v,buffer.ptr,i,_SC("interactive console"),SQTrue))){
				sq_pushroottable(v);
				if(SQ_SUCCEEDED(sq_call(v,1,retval,SQTrue)) &&	retval){
					scprintf(_SC("\n"));
					sq_pushroottable(v);
					sq_pushstring(v,_SC("print"),-1);
					sq_get(v,-2);
					sq_pushroottable(v);
					sq_push(v,-4);
					sq_call(v,2,SQFalse,SQTrue);
					retval=0;
					scprintf(_SC("\n"));
				}
			}
			
			sq_settop(v,oldtop);
		}
	}
}

